# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the Schema classes.

A schema may be one of:
    A record, mapping field names to field value data;
    An error, equivalent to a record;
    An enum, containing one of a small set of symbols;
    An array of values, all of the same schema;
    A map containing string/value pairs, each of a declared schema;
    A union of other schemas;
    A fixed sized binary object;
    A unicode string;
    A sequence of bytes;
    A 32-bit signed int;
    A 64-bit signed long;
    A 32-bit floating-point float;
    A 64-bit floating-point double;
    A boolean; or
    Null.
"""
try:
    import json
except ImportError:
    import simplejson as json

# Python3
import six

#
# Constants
#

PRIMITIVE_TYPES = (
    'null',
    'boolean',
    'string',
    'bytes',
    'int',
    'long',
    'float',
    'double',
)

NAMED_TYPES = (
    'fixed',
    'enum',
    'record',
    'error',
)

VALID_TYPES = PRIMITIVE_TYPES + NAMED_TYPES + (
    'array',
    'map',
    'union',
    'request',
    'error_union'
)

SCHEMA_RESERVED_PROPS = (
    'type',
    'name',
    'namespace',
    'fields',     # Record
    'items',      # Array
    'size',       # Fixed
    'symbols',    # Enum
    'values',     # Map
    'doc',
)

FIELD_RESERVED_PROPS = (
    'default',
    'name',
    'doc',
    'order',
    'type',
)

VALID_FIELD_SORT_ORDERS = (
    'ascending',
    'descending',
    'ignore',
)

#
# Exceptions
#


class AvroException(Exception):
    pass


class SchemaParseException(AvroException):
    pass

#
# Base Classes
#


class Schema(object):
    """Base class for all Schema classes."""
    def __init__(self, type, other_props=None):
        # Ensure valid ctor args
        if not isinstance(type, six.string_types):
            fail_msg = 'Schema type must be a string.'
            raise SchemaParseException(fail_msg)
        elif type not in VALID_TYPES:
            fail_msg = '%s is not a valid type.' % type
            raise SchemaParseException(fail_msg)

        # add members
        if not hasattr(self, '_props'):
            self._props = {}
        self.set_prop('type', type)
        self.type = type
        self._props.update(other_props or {})

    # Read-only properties dict. Printing schemas
    # creates JSON properties directly from this dict.
    props = property(lambda self: self._props)

    # Read-only property dict. Non-reserved properties
    other_props = property(lambda self: get_other_props(self._props, SCHEMA_RESERVED_PROPS),
                                                 doc="dictionary of non-reserved properties")

    # utility functions to manipulate properties dict
    def get_prop(self, key):
        return self._props.get(key)

    def set_prop(self, key, value):
        self._props[key] = value

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self, names):
        """
        Converts the schema object into its AVRO specification representation.

        Schema types that have names (records, enums, and fixed) must
        be aware of not re-defining schemas that are already listed
        in the parameter names.
        """
        raise Exception("Must be implemented by subclasses.")


class Name(object):
    """Class to describe Avro name."""

    def __init__(self, name_attr, space_attr, default_space):
        """
        Formulate full name according to the specification.

        @arg name_attr: name value read in schema or None.
        @arg space_attr: namespace value read in schema or None.
        @ard default_space: the current default space or None.
        """
        # Ensure valid ctor args
        if not (isinstance(name_attr, six.string_types) or (name_attr is None)):
            fail_msg = 'Name must be non-empty string or None.'
            raise SchemaParseException(fail_msg)
        elif name_attr == "":
            fail_msg = 'Name must be non-empty string or None.'
            raise SchemaParseException(fail_msg)

        if not (isinstance(space_attr, six.string_types) or (space_attr is None)):
            fail_msg = 'Space must be non-empty string or None.'
            raise SchemaParseException(fail_msg)
        elif name_attr == "":
            fail_msg = 'Space must be non-empty string or None.'
            raise SchemaParseException(fail_msg)

        if not (isinstance(default_space, six.string_types) or (default_space is None)):
            fail_msg = 'Default space must be non-empty string or None.'
            raise SchemaParseException(fail_msg)
        elif name_attr == "":
            fail_msg = 'Default must be non-empty string or None.'
            raise SchemaParseException(fail_msg)

        self._full = None

        if name_attr is None or name_attr == "":
            return

        if (name_attr.find('.') < 0):
            if (space_attr is not None) and (space_attr != ""):
                self._full = "%s.%s" % (space_attr, name_attr)
            else:
                if (default_space is not None) and (default_space != ""):
                    self._full = "%s.%s" % (default_space, name_attr)
                else:
                    self._full = name_attr
        else:
                self._full = name_attr

    def __eq__(self, other):
        if not isinstance(other, Name):
                return False
        return (self.fullname == other.fullname)

    fullname = property(lambda self: self._full)

    def get_space(self):
        """Back out a namespace from full name."""
        if self._full is None:
                return None

        if (self._full.find('.') > 0):
            return self._full.rsplit(".", 1)[0]
        else:
            return ""


class Names(object):
    """Track name set and default namespace during parsing."""
    def __init__(self, default_namespace=None):
            self.names = {}
            self.default_namespace = default_namespace

    def has_name(self, name_attr, space_attr):
            test = Name(name_attr, space_attr, self.default_namespace).fullname
            return test in self.names

    def get_name(self, name_attr, space_attr):
            test = Name(name_attr, space_attr, self.default_namespace).fullname
            if test not in self.names:
                    return None
            return self.names[test]

    def prune_namespace(self, properties):
        """given a properties, return properties with namespace removed if
        it matches the own default namespace"""
        if self.default_namespace is None:
            # I have no default -- no change
            return properties
        if 'namespace' not in properties:
            # he has no namespace - no change
            return properties
        if properties['namespace'] != self.default_namespace:
            # we're different - leave his stuff alone
            return properties
        # we each have a namespace and it's redundant. delete his.
        prunable = properties.copy()
        del(prunable['namespace'])
        return prunable

    def add_name(self, name_attr, space_attr, new_schema):
        """
        Add a new schema object to the name set.

            @arg name_attr: name value read in schema
            @arg space_attr: namespace value read in schema.

            @return: the Name that was just added.
        """
        to_add = Name(name_attr, space_attr, self.default_namespace)

        if to_add.fullname in VALID_TYPES:
            fail_msg = '%s is a reserved type name.' % to_add.fullname
            raise SchemaParseException(fail_msg)
        elif to_add.fullname in self.names:
            fail_msg = 'The name "%s" is already in use.' % to_add.fullname
            raise SchemaParseException(fail_msg)

        self.names[to_add.fullname] = new_schema
        return to_add


class NamedSchema(Schema):
    """Named Schemas specified in NAMED_TYPES."""
    def __init__(self, type, name, namespace=None, names=None, other_props=None):
        # Ensure valid ctor args
        if not name:
            fail_msg = 'Named Schemas must have a non-empty name.'
            raise SchemaParseException(fail_msg)
        elif not isinstance(name, six.string_types):
            fail_msg = 'The name property must be a string.'
            raise SchemaParseException(fail_msg)
        elif namespace is not None and not isinstance(namespace, six.string_types):
            fail_msg = 'The namespace property must be a string.'
            raise SchemaParseException(fail_msg)

        # Call parent ctor
        Schema.__init__(self, type, other_props)

        # Add class members
        new_name = names.add_name(name, namespace, self)

        # Store name and namespace as they were read in origin schema
        self.set_prop('name', name)
        if namespace is not None:
            self.set_prop('namespace', new_name.get_space())

        # Store full name as calculated from name, namespace
        self._fullname = new_name.fullname

    def name_ref(self, names):
        if self.namespace == names.default_namespace:
            return self.name
        else:
            return self.fullname

    # read-only properties
    name = property(lambda self: self.get_prop('name'))
    namespace = property(lambda self: self.get_prop('namespace'))
    fullname = property(lambda self: self._fullname)


class Field(object):
    def __init__(self, type, name, has_default, default=None,
                             order=None, names=None, doc=None, other_props=None):
        # Ensure valid ctor args
        if not name:
            fail_msg = 'Fields must have a non-empty name.'
            raise SchemaParseException(fail_msg)
        elif not isinstance(name, six.string_types):
            fail_msg = 'The name property must be a string.'
            raise SchemaParseException(fail_msg)
        elif order is not None and order not in VALID_FIELD_SORT_ORDERS:
            fail_msg = 'The order property %s is not valid.' % order
            raise SchemaParseException(fail_msg)

        # add members
        self._props = {}
        self._has_default = has_default
        self._props.update(other_props or {})

        if (isinstance(type, six.string_types) and names is not None
                and names.has_name(type, None)):
            type_schema = names.get_name(type, None)
        else:
            try:
                type_schema = make_avsc_object(type, names)
            except Exception as e:
                fail_msg = 'Type property "%s" not a valid Avro schema: %s' % (type, e)
                raise SchemaParseException(fail_msg)
        self.set_prop('type', type_schema)
        self.set_prop('name', name)
        self.type = type_schema
        self.name = name
        # TODO(hammer): check to ensure default is valid
        if has_default:
            self.set_prop('default', default)
        if order is not None:
            self.set_prop('order', order)
        if doc is not None:
            self.set_prop('doc', doc)

    # read-only properties
    default = property(lambda self: self.get_prop('default'))
    has_default = property(lambda self: self._has_default)
    order = property(lambda self: self.get_prop('order'))
    doc = property(lambda self: self.get_prop('doc'))
    props = property(lambda self: self._props)

    # Read-only property dict. Non-reserved properties
    other_props = property(lambda self: get_other_props(self._props, FIELD_RESERVED_PROPS),
                                                 doc="dictionary of non-reserved properties")

# utility functions to manipulate properties dict
    def get_prop(self, key):
        return self._props.get(key)

    def set_prop(self, key, value):
        self._props[key] = value

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = self.props.copy()
        to_dump['type'] = self.type.to_json(names)
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))

#
# Primitive Types
#


class PrimitiveSchema(Schema):
    """Valid primitive types are in PRIMITIVE_TYPES."""
    def __init__(self, type, other_props=None):
        # Ensure valid ctor args
        if type not in PRIMITIVE_TYPES:
            raise AvroException("%s is not a valid primitive type." % type)

        # Call parent ctor
        Schema.__init__(self, type, other_props=other_props)

        self.fullname = type

    def to_json(self, names=None):
        if len(self.props) == 1:
            return self.fullname
        else:
            return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# Complex Types (non-recursive)
#


class FixedSchema(NamedSchema):
    def __init__(self, name, namespace, size, names=None, other_props=None):
        # Ensure valid ctor args
        if not isinstance(size, int):
            fail_msg = 'Fixed Schema requires a valid integer for size property.'
            raise AvroException(fail_msg)

        # Call parent ctor
        NamedSchema.__init__(self, 'fixed', name, namespace, names, other_props)

        # Add class members
        self.set_prop('size', size)

    # read-only properties
    size = property(lambda self: self.get_prop('size'))

    def to_json(self, names=None):
        if names is None:
            names = Names()
        if self.fullname in names.names:
            return self.name_ref(names)
        else:
            names.names[self.fullname] = self
            return names.prune_namespace(self.props)

    def __eq__(self, that):
        return self.props == that.props


class EnumSchema(NamedSchema):
    def __init__(self, name, namespace, symbols, names=None, doc=None, other_props=None):
        # Ensure valid ctor args
        if not isinstance(symbols, list):
            fail_msg = 'Enum Schema requires a JSON array for the symbols property.'
            raise AvroException(fail_msg)
        elif False in [isinstance(s, six.string_types) for s in symbols]:
            fail_msg = 'Enum Schema requires all symbols to be JSON strings.'
            raise AvroException(fail_msg)
        elif len(set(symbols)) < len(symbols):
            fail_msg = 'Duplicate symbol: %s' % symbols
            raise AvroException(fail_msg)

        # Call parent ctor
        NamedSchema.__init__(self, 'enum', name, namespace, names, other_props)

        # Add class members
        self.set_prop('symbols', symbols)
        if doc is not None:
            self.set_prop('doc', doc)

    # read-only properties
    symbols = property(lambda self: self.get_prop('symbols'))
    doc = property(lambda self: self.get_prop('doc'))

    def to_json(self, names=None):
        if names is None:
            names = Names()
        if self.fullname in names.names:
            return self.name_ref(names)
        else:
            names.names[self.fullname] = self
            return names.prune_namespace(self.props)

    def __eq__(self, that):
        return self.props == that.props

#
# Complex Types (recursive)
#


class ArraySchema(Schema):
    def __init__(self, items, names=None, other_props=None):
        # Call parent ctor
        Schema.__init__(self, 'array', other_props)
        # Add class members

        if isinstance(items, six.string_types) and names.has_name(items, None):
            items_schema = names.get_name(items, None)
        else:
            try:
                items_schema = make_avsc_object(items, names)
            except SchemaParseException as e:
                fail_msg = 'Items schema (%s) not a valid Avro schema: %s (known names: %s)' % (items, e, names.names.keys())
                raise SchemaParseException(fail_msg)

        self.set_prop('items', items_schema)

    # read-only properties
    items = property(lambda self: self.get_prop('items'))

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = self.props.copy()
        item_schema = self.get_prop('items')
        to_dump['items'] = item_schema.to_json(names)
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class MapSchema(Schema):
    def __init__(self, values, names=None, other_props=None):
        # Call parent ctor
        Schema.__init__(self, 'map', other_props)

        # Add class members
        if isinstance(values, six.string_types) and names.has_name(values, None):
            values_schema = names.get_name(values, None)
        else:
            try:
                values_schema = make_avsc_object(values, names)
            except:
                fail_msg = 'Values schema not a valid Avro schema.'
                raise SchemaParseException(fail_msg)

        self.set_prop('values', values_schema)

    # read-only properties
    values = property(lambda self: self.get_prop('values'))

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = self.props.copy()
        to_dump['values'] = self.get_prop('values').to_json(names)
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class UnionSchema(Schema):
    """
    names is a dictionary of schema objects
    """
    def __init__(self, schemas, names=None):
        # Ensure valid ctor args
        if not isinstance(schemas, list):
            fail_msg = 'Union schema requires a list of schemas.'
            raise SchemaParseException(fail_msg)

        # Call parent ctor
        Schema.__init__(self, 'union')

        # Add class members
        schema_objects = []
        for schema in schemas:
            if isinstance(schema, six.string_types) and names.has_name(schema, None):
                new_schema = names.get_name(schema, None)
            else:
                try:
                    new_schema = make_avsc_object(schema, names)
                except Exception as e:
                    raise SchemaParseException('Union item must be a valid Avro schema: %s' % str(e))
            # check the new schema
            if (new_schema.type in VALID_TYPES and new_schema.type not in NAMED_TYPES
                    and new_schema.type in [schema.type for schema in schema_objects]):
                raise SchemaParseException('%s type already in Union' % new_schema.type)
            elif new_schema.type == 'union':
                raise SchemaParseException('Unions cannot contain other unions.')
            else:
                schema_objects.append(new_schema)
        self._schemas = schema_objects

    # read-only properties
    schemas = property(lambda self: self._schemas)

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = []
        for schema in self.schemas:
            to_dump.append(schema.to_json(names))
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class ErrorUnionSchema(UnionSchema):
    def __init__(self, schemas, names=None):
        # Prepend "string" to handle system errors
        UnionSchema.__init__(self, ['string'] + schemas, names)

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = []
        for schema in self.schemas:
            # Don't print the system error schema
            if schema.type == 'string': continue
            to_dump.append(schema.to_json(names))
        return to_dump


class RecordSchema(NamedSchema):
    @staticmethod
    def make_field_objects(field_data, names):
        """We're going to need to make message parameters too."""
        field_objects = []
        field_names = []
        for i, field in enumerate(field_data):
            if hasattr(field, 'get') and callable(field.get):
                type = field.get('type')
                name = field.get('name')

                # null values can have a default value of None
                has_default = False
                default = None
                if 'default' in field:
                    has_default = True
                    default = field.get('default')

                order = field.get('order')
                doc = field.get('doc')
                other_props = get_other_props(field, FIELD_RESERVED_PROPS)
                new_field = Field(type, name, has_default, default, order, names, doc,
                                                 other_props)
                # make sure field name has not been used yet
                if new_field.name in field_names:
                    fail_msg = 'Field name %s already in use.' % new_field.name
                    raise SchemaParseException(fail_msg)
                field_names.append(new_field.name)
            else:
                raise SchemaParseException('Not a valid field: %s' % field)
            field_objects.append(new_field)
        return field_objects

    def __init__(self, name, namespace, fields, names=None, schema_type='record',
                             doc=None, other_props=None):
        # Ensure valid ctor args
        if fields is None:
            fail_msg = 'Record schema requires a non-empty fields property.'
            raise SchemaParseException(fail_msg)
        elif not isinstance(fields, list):
            fail_msg = 'Fields property must be a list of Avro schemas.'
            raise SchemaParseException(fail_msg)

        # Call parent ctor (adds own name to namespace, too)
        if schema_type == 'request':
            Schema.__init__(self, schema_type, other_props)
        else:
            NamedSchema.__init__(self, schema_type, name, namespace, names,
                                                     other_props)

        if schema_type == 'record':
            old_default = names.default_namespace
            names.default_namespace = Name(name, namespace,
                                                                         names.default_namespace).get_space()

        # Add class members
        field_objects = RecordSchema.make_field_objects(fields, names)
        self.set_prop('fields', field_objects)
        if doc is not None:
            self.set_prop('doc', doc)

        if schema_type == 'record':
            names.default_namespace = old_default

    # read-only properties
    fields = property(lambda self: self.get_prop('fields'))
    doc = property(lambda self: self.get_prop('doc'))

    @property
    def fields_dict(self):
        fields_dict = {}
        for field in self.fields:
            fields_dict[field.name] = field
        return fields_dict

    def to_json(self, names=None):
        if names is None:
            names = Names()
        # Request records don't have names
        if self.type == 'request':
            return [f.to_json(names) for f in self.fields]

        if self.fullname in names.names:
            return self.name_ref(names)
        else:
            names.names[self.fullname] = self

        to_dump = names.prune_namespace(self.props.copy())
        to_dump['fields'] = [f.to_json(names) for f in self.fields]
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))

#
# Module Methods
#


def get_other_props(all_props, reserved_props):
    """
    Retrieve the non-reserved properties from a dictionary of properties
    @args reserved_props: The set of reserved properties to exclude
    """
    if hasattr(all_props, 'items') and callable(all_props.items):
        return dict([(k, v) for (k, v) in all_props.items() if k not in
                                 reserved_props])


def make_avsc_object(json_data, names=None):
    """
    Build Avro Schema from data parsed out of JSON string.

    @arg names: A Name object (tracks seen names and default space)
    """
    if names is None:
        names = Names()

    # JSON object (non-union)
    if hasattr(json_data, 'get') and callable(json_data.get):
        type = json_data.get('type')
        other_props = get_other_props(json_data, SCHEMA_RESERVED_PROPS)
        if type in PRIMITIVE_TYPES:
            return PrimitiveSchema(type, other_props)
        elif type in NAMED_TYPES:
            name = json_data.get('name')
            namespace = json_data.get('namespace', names.default_namespace)
            if type == 'fixed':
                size = json_data.get('size')
                return FixedSchema(name, namespace, size, names, other_props)
            elif type == 'enum':
                symbols = json_data.get('symbols')
                doc = json_data.get('doc')
                return EnumSchema(name, namespace, symbols, names, doc, other_props)
            elif type in ['record', 'error']:
                fields = json_data.get('fields')
                doc = json_data.get('doc')
                return RecordSchema(name, namespace, fields, names, type, doc, other_props)
            else:
                raise SchemaParseException('Unknown Named Type: %s' % type)
        elif type in VALID_TYPES:
            if type == 'array':
                items = json_data.get('items')
                return ArraySchema(items, names, other_props)
            elif type == 'map':
                values = json_data.get('values')
                return MapSchema(values, names, other_props)
            elif type == 'error_union':
                declared_errors = json_data.get('declared_errors')
                return ErrorUnionSchema(declared_errors, names)
            else:
                raise SchemaParseException('Unknown Valid Type: %s' % type)
        elif type is None:
            raise SchemaParseException('No "type" property: %s' % json_data)
        else:
            raise SchemaParseException('Undefined type: %s' % type)
    # JSON array (union)
    elif isinstance(json_data, list):
        return UnionSchema(json_data, names)
    # JSON string (primitive)
    elif json_data in PRIMITIVE_TYPES:
        return PrimitiveSchema(json_data)
    # not for us!
    else:
        fail_msg = "Could not make an Avro Schema object from %s." % json_data
        raise SchemaParseException(fail_msg)


# TODO(hammer): make method for reading from a file?
def parse(json_string):
    """Constructs the Schema from the JSON text."""
    # parse the JSON
    try:
        json_data = json.loads(json_string)
    except Exception as e:
        import sys
        raise SchemaParseException('Error parsing JSON: %s, error = %s'
                                                        % (json_string, e)) # , None, sys.exc_info()[2]

    # Initialize the names object
    names = Names()

    # construct the Avro Schema object
    return make_avsc_object(json_data, names)



#! /usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from sys import version_info
from distutils.extension import Extension

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

install_requires = ['six>=1.10.0']
if version_info[:2] <= (2, 5):
    install_requires.append('simplejson >= 2.0.9')

ext = '.pyx' if USE_CYTHON else '.c'
extensions = [Extension("spavro.fast_binary", sources=["src/spavro/fast_binary" + ext])]

if USE_CYTHON:
    extensions = cythonize(extensions)

setup(
  name='spavro',
  version='1.1.3',
  packages=['spavro'],
  package_dir={'': 'src'},
  # scripts=["./scripts/avro"],
  # Project uses simplejson, so ensure that it gets installed or upgraded
  # on the target machine
  install_requires=install_requires,
  # spavro C extensions
  ext_modules=extensions,
  # metadata for upload to PyPI
  author='Michael Kowalchik',
  author_email='mikepk@pluralsight.com',
  description='Spavro is a (sp)eedier avro library -- Spavro is a fork of the official Apache AVRO python 2 implementation with the goal of greatly improving data read deserialization and write serialization performance.',
  license='Apache License 2.0',
  keywords='avro serialization rpc data',
  url='http://github.com/pluralsight/spavro',
  extras_require={
    'snappy': ['python-snappy'],
    'test': ['pytest>=3.1.1'],
  },
  classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.5",
            "Topic :: Software Development :: Libraries",
            "Topic :: System :: Networking",
            "Operating System :: OS Independent",
        ]
)



# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Support for inter-process calls.
"""
import six
from six.moves import http_client as httplib
from six import BytesIO as StringIO
from spavro import io
from spavro import protocol
from spavro import schema

#
# Constants
#

# Handshake schema is pulled in during build
with open('./ipc/HandshakeRequest.avsc', 'rb') as handshake_request:
    HANDSHAKE_REQUEST_SCHEMA = schema.parse(handshake_request.read().decode('utf-8'))
# """
# """)
with open('./ipc/HandshakeResponse.avsc', 'rb') as handshake_response:
    HANDSHAKE_RESPONSE_SCHEMA = schema.parse(handshake_response.read().decode('utf-8'))

# HANDSHAKE_RESPONSE_SCHEMA = schema.parse("""
# @HANDSHAKE_RESPONSE_SCHEMA@
# """)

HANDSHAKE_REQUESTOR_WRITER = io.DatumWriter(HANDSHAKE_REQUEST_SCHEMA)
HANDSHAKE_REQUESTOR_READER = io.DatumReader(HANDSHAKE_RESPONSE_SCHEMA)
HANDSHAKE_RESPONDER_WRITER = io.DatumWriter(HANDSHAKE_RESPONSE_SCHEMA)
HANDSHAKE_RESPONDER_READER = io.DatumReader(HANDSHAKE_REQUEST_SCHEMA)

META_SCHEMA = schema.parse('{"type": "map", "values": "bytes"}')
META_WRITER = io.DatumWriter(META_SCHEMA)
META_READER = io.DatumReader(META_SCHEMA)

SYSTEM_ERROR_SCHEMA = schema.parse('["string"]')

# protocol cache
REMOTE_HASHES = {}
REMOTE_PROTOCOLS = {}

BIG_ENDIAN_INT_STRUCT = io.struct_class('!I')
BUFFER_HEADER_LENGTH = 4
BUFFER_SIZE = 8192

#
# Exceptions
#

class AvroRemoteException(schema.AvroException):
  """
  Raised when an error message is sent by an Avro requestor or responder.
  """
  def __init__(self, fail_msg=None):
    schema.AvroException.__init__(self, fail_msg)

class ConnectionClosedException(schema.AvroException):
  pass

#
# Base IPC Classes (Requestor/Responder)
#

class BaseRequestor(object):
  """Base class for the client side of a protocol interaction."""
  def __init__(self, local_protocol, transceiver):
    self._local_protocol = local_protocol
    self._transceiver = transceiver
    self._remote_protocol = None
    self._remote_hash = None
    self._send_protocol = None

  # read-only properties
  local_protocol = property(lambda self: self._local_protocol)
  transceiver = property(lambda self: self._transceiver)

  # read/write properties
  def set_remote_protocol(self, new_remote_protocol):
    self._remote_protocol = new_remote_protocol
    REMOTE_PROTOCOLS[self.transceiver.remote_name] = self.remote_protocol
  remote_protocol = property(lambda self: self._remote_protocol,
                             set_remote_protocol)

  def set_remote_hash(self, new_remote_hash):
    self._remote_hash = new_remote_hash
    REMOTE_HASHES[self.transceiver.remote_name] = self.remote_hash
  remote_hash = property(lambda self: self._remote_hash, set_remote_hash)

  def set_send_protocol(self, new_send_protocol):
    self._send_protocol = new_send_protocol
  send_protocol = property(lambda self: self._send_protocol, set_send_protocol)

  def request(self, message_name, request_datum):
    """
    Writes a request message and reads a response or error message.
    """
    # build handshake and call request
    buffer_writer = StringIO()
    buffer_encoder = io.BinaryEncoder(buffer_writer)
    self.write_handshake_request(buffer_encoder)
    self.write_call_request(message_name, request_datum, buffer_encoder)

    # send the handshake and call request; block until call response
    call_request = buffer_writer.getvalue()
    return self.issue_request(call_request, message_name, request_datum)

  def write_handshake_request(self, encoder):
    local_hash = self.local_protocol.md5
    remote_name = self.transceiver.remote_name
    remote_hash = REMOTE_HASHES.get(remote_name)
    if remote_hash is None:
      remote_hash = local_hash
      self.remote_protocol = self.local_protocol
    request_datum = {}
    request_datum['clientHash'] = local_hash
    request_datum['serverHash'] = remote_hash
    if self.send_protocol:
      request_datum['clientProtocol'] = str(self.local_protocol)
    HANDSHAKE_REQUESTOR_WRITER.write(request_datum, encoder)

  def write_call_request(self, message_name, request_datum, encoder):
    """
    The format of a call request is:
      * request metadata, a map with values of type bytes
      * the message name, an Avro string, followed by
      * the message parameters. Parameters are serialized according to
        the message's request declaration.
    """
    # request metadata (not yet implemented)
    request_metadata = {}
    META_WRITER.write(request_metadata, encoder)

    # message name
    message = self.local_protocol.messages.get(message_name)
    if message is None:
      raise schema.AvroException('Unknown message: %s' % message_name)
    encoder.write_utf8(message.name)

    # message parameters
    self.write_request(message.request, request_datum, encoder)

  def write_request(self, request_schema, request_datum, encoder):
    datum_writer = io.DatumWriter(request_schema)
    datum_writer.write(request_datum, encoder)

  def read_handshake_response(self, decoder):
    handshake_response = HANDSHAKE_REQUESTOR_READER.read(decoder)
    match = handshake_response.get('match')
    if match == 'BOTH':
      self.send_protocol = False
      return True
    elif match == 'CLIENT':
      if self.send_protocol:
        raise schema.AvroException('Handshake failure.')
      self.remote_protocol = protocol.parse(
                             handshake_response.get('serverProtocol'))
      self.remote_hash = handshake_response.get('serverHash')
      self.send_protocol = False
      return True
    elif match == 'NONE':
      if self.send_protocol:
        raise schema.AvroException('Handshake failure.')
      self.remote_protocol = protocol.parse(
                             handshake_response.get('serverProtocol'))
      self.remote_hash = handshake_response.get('serverHash')
      self.send_protocol = True
      return False
    else:
      raise schema.AvroException('Unexpected match: %s' % match)

  def read_call_response(self, message_name, decoder):
    """
    The format of a call response is:
      * response metadata, a map with values of type bytes
      * a one-byte error flag boolean, followed by either:
        o if the error flag is false,
          the message response, serialized per the message's response schema.
        o if the error flag is true, 
          the error, serialized per the message's error union schema.
    """
    # response metadata
    response_metadata = META_READER.read(decoder)

    # remote response schema
    remote_message_schema = self.remote_protocol.messages.get(message_name)
    if remote_message_schema is None:
      raise schema.AvroException('Unknown remote message: %s' % message_name)

    # local response schema
    local_message_schema = self.local_protocol.messages.get(message_name)
    if local_message_schema is None:
      raise schema.AvroException('Unknown local message: %s' % message_name)

    # error flag
    if not decoder.read_boolean():
      writers_schema = remote_message_schema.response
      readers_schema = local_message_schema.response
      return self.read_response(writers_schema, readers_schema, decoder)
    else:
      writers_schema = remote_message_schema.errors
      readers_schema = local_message_schema.errors
      raise self.read_error(writers_schema, readers_schema, decoder)

  def read_response(self, writers_schema, readers_schema, decoder):
    datum_reader = io.DatumReader(writers_schema, readers_schema)
    result = datum_reader.read(decoder)
    return result

  def read_error(self, writers_schema, readers_schema, decoder):
    datum_reader = io.DatumReader(writers_schema, readers_schema)
    return AvroRemoteException(datum_reader.read(decoder))

class Requestor(BaseRequestor):

  def issue_request(self, call_request, message_name, request_datum):
    call_response = self.transceiver.transceive(call_request)

    # process the handshake and call response
    buffer_decoder = io.BinaryDecoder(StringIO(call_response))
    call_response_exists = self.read_handshake_response(buffer_decoder)
    if call_response_exists:
      return self.read_call_response(message_name, buffer_decoder)
    else:
      return self.request(message_name, request_datum)

class Responder(object):
  """Base class for the server side of a protocol interaction."""
  def __init__(self, local_protocol):
    self._local_protocol = local_protocol
    self._local_hash = self.local_protocol.md5
    self._protocol_cache = {}
    self.set_protocol_cache(self.local_hash, self.local_protocol)

  # read-only properties
  local_protocol = property(lambda self: self._local_protocol)
  local_hash = property(lambda self: self._local_hash)
  protocol_cache = property(lambda self: self._protocol_cache)

  # utility functions to manipulate protocol cache
  def get_protocol_cache(self, hash):
    return self.protocol_cache.get(hash)
  def set_protocol_cache(self, hash, protocol):
    self.protocol_cache[hash] = protocol

  def respond(self, call_request):
    """
    Called by a server to deserialize a request, compute and serialize
    a response or error. Compare to 'handle()' in Thrift.
    """
    buffer_reader = StringIO(call_request)
    buffer_decoder = io.BinaryDecoder(buffer_reader)
    buffer_writer = StringIO()
    buffer_encoder = io.BinaryEncoder(buffer_writer)
    error = None
    response_metadata = {}
    
    try:
      remote_protocol = self.process_handshake(buffer_decoder, buffer_encoder)
      # handshake failure
      if remote_protocol is None:  
        return buffer_writer.getvalue()

      # read request using remote protocol
      request_metadata = META_READER.read(buffer_decoder)
      remote_message_name = buffer_decoder.read_utf8()

      # get remote and local request schemas so we can do
      # schema resolution (one fine day)
      remote_message = remote_protocol.messages.get(remote_message_name)
      if remote_message is None:
        fail_msg = 'Unknown remote message: %s' % remote_message_name
        raise schema.AvroException(fail_msg)
      local_message = self.local_protocol.messages.get(remote_message_name)
      if local_message is None:
        fail_msg = 'Unknown local message: %s' % remote_message_name
        raise schema.AvroException(fail_msg)
      writers_schema = remote_message.request
      readers_schema = local_message.request
      request = self.read_request(writers_schema, readers_schema,
                                  buffer_decoder)

      # perform server logic
      try:
        response = self.invoke(local_message, request)
      except AvroRemoteException as e:
        error = e
      except Exception as e:
        error = AvroRemoteException(str(e))

      # write response using local protocol
      META_WRITER.write(response_metadata, buffer_encoder)
      buffer_encoder.write_boolean(error is not None)
      if error is None:
        writers_schema = local_message.response
        self.write_response(writers_schema, response, buffer_encoder)
      else:
        writers_schema = local_message.errors
        self.write_error(writers_schema, error, buffer_encoder)
    except schema.AvroException as e:
      error = AvroRemoteException(str(e))
      buffer_encoder = io.BinaryEncoder(StringIO())
      META_WRITER.write(response_metadata, buffer_encoder)
      buffer_encoder.write_boolean(True)
      self.write_error(SYSTEM_ERROR_SCHEMA, error, buffer_encoder)
    return buffer_writer.getvalue()

  def process_handshake(self, decoder, encoder):
    handshake_request = HANDSHAKE_RESPONDER_READER.read(decoder)
    handshake_response = {}

    # determine the remote protocol
    client_hash = handshake_request.get('clientHash')
    client_protocol = handshake_request.get('clientProtocol')
    remote_protocol = self.get_protocol_cache(client_hash)
    if remote_protocol is None and client_protocol is not None:
      remote_protocol = protocol.parse(client_protocol)
      self.set_protocol_cache(client_hash, remote_protocol)

    # evaluate remote's guess of the local protocol
    server_hash = handshake_request.get('serverHash')
    if self.local_hash == server_hash:
      if remote_protocol is None:
        handshake_response['match'] = 'NONE'
      else:
        handshake_response['match'] = 'BOTH'
    else:
      if remote_protocol is None:
        handshake_response['match'] = 'NONE'
      else:
        handshake_response['match'] = 'CLIENT'

    if handshake_response['match'] != 'BOTH':
      handshake_response['serverProtocol'] = str(self.local_protocol)
      handshake_response['serverHash'] = self.local_hash

    HANDSHAKE_RESPONDER_WRITER.write(handshake_response, encoder)
    return remote_protocol

  def invoke(self, local_message, request):
    """
    Aactual work done by server: cf. handler in thrift.
    """
    pass

  def read_request(self, writers_schema, readers_schema, decoder):
    datum_reader = io.DatumReader(writers_schema, readers_schema)
    return datum_reader.read(decoder)

  def write_response(self, writers_schema, response_datum, encoder):
    datum_writer = io.DatumWriter(writers_schema)
    datum_writer.write(response_datum, encoder)

  def write_error(self, writers_schema, error_exception, encoder):
    datum_writer = io.DatumWriter(writers_schema)
    datum_writer.write(str(error_exception), encoder)

#
# Utility classes
#

class FramedReader(object):
  """Wrapper around a file-like object to read framed data."""
  def __init__(self, reader):
    self._reader = reader

  # read-only properties
  reader = property(lambda self: self._reader)

  def read_framed_message(self):
    message = []
    while True:
      buffer = StringIO()
      buffer_length = self._read_buffer_length()
      if buffer_length == 0:
        return ''.join(message)
      while buffer.tell() < buffer_length:
        chunk = self.reader.read(buffer_length - buffer.tell())
        if chunk == '':
          raise ConnectionClosedException("Reader read 0 bytes.")
        buffer.write(chunk)
      message.append(buffer.getvalue())

  def _read_buffer_length(self):
    read = self.reader.read(BUFFER_HEADER_LENGTH)
    if read == '':
      raise ConnectionClosedException("Reader read 0 bytes.")
    return BIG_ENDIAN_INT_STRUCT.unpack(read)[0]

class FramedWriter(object):
  """Wrapper around a file-like object to write framed data."""
  def __init__(self, writer):
    self._writer = writer

  # read-only properties
  writer = property(lambda self: self._writer)

  def write_framed_message(self, message):
    message_length = len(message)
    total_bytes_sent = 0
    while message_length - total_bytes_sent > 0:
      if message_length - total_bytes_sent > BUFFER_SIZE:
        buffer_length = BUFFER_SIZE
      else:
        buffer_length = message_length - total_bytes_sent
      self.write_buffer(message[total_bytes_sent:
                                (total_bytes_sent + buffer_length)])
      total_bytes_sent += buffer_length
    # A message is always terminated by a zero-length buffer.
    self.write_buffer_length(0)

  def write_buffer(self, chunk):
    buffer_length = len(chunk)
    self.write_buffer_length(buffer_length)
    self.writer.write(chunk)

  def write_buffer_length(self, n):
    self.writer.write(BIG_ENDIAN_INT_STRUCT.pack(n))

#
# Transceiver Implementations
#

class HTTPTransceiver(object):
  """
  A simple HTTP-based transceiver implementation.
  Useful for clients but not for servers
  """
  def __init__(self, host, port, req_resource='/'):
    self.req_resource = req_resource
    self.conn = httplib.HTTPConnection(host, port)
    self.conn.connect()

  # read-only properties
  sock = property(lambda self: self.conn.sock)
  remote_name = property(lambda self: self.sock.getsockname())

  # read/write properties
  def set_conn(self, new_conn):
    self._conn = new_conn
  conn = property(lambda self: self._conn, set_conn)
  req_resource = '/'

  def transceive(self, request):
    self.write_framed_message(request)
    result = self.read_framed_message()
    return result

  def read_framed_message(self):
    response = self.conn.getresponse()
    response_reader = FramedReader(response)
    framed_message = response_reader.read_framed_message()
    response.read()    # ensure we're ready for subsequent requests
    return framed_message

  def write_framed_message(self, message):
    req_method = 'POST'
    req_headers = {'Content-Type': 'avro/binary'}

    req_body_buffer = FramedWriter(StringIO())
    req_body_buffer.write_framed_message(message)
    req_body = req_body_buffer.writer.getvalue()

    self.conn.request(req_method, self.req_resource, req_body, req_headers)

  def close(self):
    self.conn.close()

#
# Server Implementations (none yet)
#

